import torch
from torch import nn
import torch.nn.functional as F
from torch.autograd import Variable
from torchvision.models import vgg19
import settings
import numpy
from itertools import combinations,product

import math
class ConvDirec(nn.Module):
    def __init__(self, inp_dim, oup_dim, kernel, dilation):
        super().__init__()
        pad = int(dilation * (kernel - 1) / 2)
        self.conv = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad, dilation=dilation)
        self.relu = nn.LeakyReLU(0.2)

    def forward(self, x, h=None):
        x = self.conv(x)
        x = self.relu(x)
        return x, None


class ConvRNN(nn.Module):
    def __init__(self, inp_dim, oup_dim, kernel, dilation):
        super().__init__()
        pad_x = int(dilation * (kernel - 1) / 2)
        self.conv_x = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad_x, dilation=dilation)

        pad_h = int((kernel - 1) / 2)
        self.conv_h = nn.Conv2d(oup_dim, oup_dim, kernel, padding=pad_h)
        self.relu = nn.LeakyReLU(0.2)

    def forward(self, x, h=None):
        if h is None:
            h = F.tanh(self.conv_x(x))
        else:
            h = F.tanh(self.conv_x(x) + self.conv_h(h))

        h = self.relu(h)
        return h, h


class ConvGRU(nn.Module):
    def __init__(self, inp_dim, oup_dim, kernel, dilation):
        super().__init__()
        pad_x = int(dilation * (kernel - 1) / 2)
        self.conv_xz = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad_x, dilation=dilation)
        self.conv_xr = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad_x, dilation=dilation)
        self.conv_xn = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad_x, dilation=dilation)

        pad_h = int((kernel - 1) / 2)
        self.conv_hz = nn.Conv2d(oup_dim, oup_dim, kernel, padding=pad_h)
        self.conv_hr = nn.Conv2d(oup_dim, oup_dim, kernel, padding=pad_h)
        self.conv_hn = nn.Conv2d(oup_dim, oup_dim, kernel, padding=pad_h)

        self.relu = nn.LeakyReLU(0.2)

    def forward(self, x, h=None):
        if h is None:
            z = F.sigmoid(self.conv_xz(x))
            f = F.tanh(self.conv_xn(x))
            h = z * f
        else:
            z = F.sigmoid(self.conv_xz(x) + self.conv_hz(h))
            r = F.sigmoid(self.conv_xr(x) + self.conv_hr(h))
            n = F.tanh(self.conv_xn(x) + self.conv_hn(r * h))
            h = (1 - z) * h + z * n

        h = self.relu(h)
        return h, h


class ConvLSTM(nn.Module):
    def __init__(self, inp_dim, oup_dim, kernel, dilation):
        super().__init__()
        pad_x = int(dilation * (kernel - 1) / 2)
        self.conv_xf = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad_x, dilation=dilation)
        self.conv_xi = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad_x, dilation=dilation)
        self.conv_xo = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad_x, dilation=dilation)
        self.conv_xj = nn.Conv2d(inp_dim, oup_dim, kernel, padding=pad_x, dilation=dilation)

        pad_h = int((kernel - 1) / 2)
        self.conv_hf = nn.Conv2d(oup_dim, oup_dim, kernel, padding=pad_h)
        self.conv_hi = nn.Conv2d(oup_dim, oup_dim, kernel, padding=pad_h)
        self.conv_ho = nn.Conv2d(oup_dim, oup_dim, kernel, padding=pad_h)
        self.conv_hj = nn.Conv2d(oup_dim, oup_dim, kernel, padding=pad_h)

        self.relu = nn.LeakyReLU(0.2)

    def forward(self, x, pair=None):
        if pair is None:
            i = F.sigmoid(self.conv_xi(x))
            o = F.sigmoid(self.conv_xo(x))
            j = F.tanh(self.conv_xj(x))
            c = i * j
            h = o * c
        else:
            h, c = pair
            f = F.sigmoid(self.conv_xf(x) + self.conv_hf(h))
            i = F.sigmoid(self.conv_xi(x) + self.conv_hi(h))
            o = F.sigmoid(self.conv_xo(x) + self.conv_ho(h))
            j = F.tanh(self.conv_xj(x) + self.conv_hj(h))
            c = f * c + i * j
            h = o * F.tanh(c)

        h = self.relu(h)
        return h, [h, c]


RecUnit = {
    'Conv': ConvDirec,
    'RNN': ConvRNN,
    'GRU': ConvGRU,
    'LSTM': ConvLSTM,
}[settings.uint]

class SKblock(nn.Module):
    def __init__(self, r = 4):
        super(SKblock, self).__init__()
        input_dim = settings.channel_derain
        self.conv1 = nn.Sequential(nn.Conv2d(input_dim, input_dim, 3, 1, 1, 1), nn.LeakyReLU(0.2))
        self.conv3 = nn.Sequential(nn.Conv2d(input_dim, input_dim, 3, 1, 3, 3), nn.LeakyReLU(0.2))
        self.conv5 = nn.Sequential(nn.Conv2d(input_dim, input_dim, 3, 1, 5, 5), nn.LeakyReLU(0.2))
        if settings.sk is True:
            self.avg_pool_1_3 = nn.AdaptiveAvgPool2d(1)
            self.fc_1_3 = nn.Sequential(
                nn.Linear(input_dim, r),
                nn.ReLU(inplace=True),
            )
            self.fc_1_3_1 = nn.Sequential(
                nn.Linear(r, input_dim),
                nn.Softmax()
            )
            self.fc_1_3_3 = nn.Sequential(
                nn.Linear(r, input_dim),
                nn.Softmax()
            )
            self.avg_pool_3_5 = nn.AdaptiveAvgPool2d(1)
            self.fc_3_5 = nn.Sequential(
                nn.Linear(input_dim, r),
                nn.ReLU(inplace=True),
            )
            self.fc_3_5_3 = nn.Sequential(
                nn.Linear(r, input_dim),
                nn.Softmax()
            )
            self.fc_3_5_5 = nn.Sequential(
                nn.Linear(r, input_dim),
                nn.Softmax()
            )
    def forward(self, x):
        b, c, h, w = x.size()
        if settings.sk is True:
            conv1 = self.conv1(x)
            conv3 = self.conv3(x)
            conv5 = self.conv5(x)
            output = conv1 + conv3 + conv5
            return output
        else:
            conv1 = self.conv1(x)
            conv3 = self.conv3(x)
            conv5 = self.conv5(x)
            conv_sum_1_3 = conv1 + conv3
            y_1_3 = self.avg_pool(conv_sum_1_3).view(b, c)
            fc_1_3 = self.fc_1_3(y_1_3).view(b, c, 1, 1)
            y_1_3_1 = self.fc_1_3_1(fc_1_3)
            y_1_3_1 = conv1 * y_1_3_1
            y_1_3_3 = self.fc_1_3_3(fc_1_3)
            y_1_3_3 = conv3 * y_1_3_3
            y_1_3 = y_1_3_1 + y_1_3_3

            conv_sum_3_5 = conv3 + conv5
            y_3_5 = self.avg_pool(conv_sum_3_5).view(b, c)
            fc_3_5 = self.fc_3_5(y_3_5).view(b, c, 1, 1)
            y_3_5_3 = self.fc_3_5_3(fc_3_5)
            y_3_5_5 = conv3 * y_3_5_3
            y_3_5_3 = self.fc_3_5_5(fc_3_5)
            y_3_5_5 = conv5 * y_3_5_5
            y_3_5 = y_3_5_3 + y_3_5_5
            output = y_1_3 + y_3_5
            return output
class Dense_block(nn.Module):
    def __init__(self,channel,num):
        super(Dense_block, self).__init__()
        self.num=num
        self.conv = nn.ModuleList()
        self.dense_cat = nn.ModuleList()
        if settings.dilation is True:
            for i in range(self.num):
                self.conv.append(nn.Sequential(nn.Conv2d(channel, channel, 3, 1, 2**i, 2**i), nn.LeakyReLU(0.2)))
            for i in range(self.num):
                self.dense_cat.append(nn.Sequential(nn.Conv2d((i+2)*channel, channel, 1, 1), nn.LeakyReLU(0.2)))
        else:
            for i in range(self.num):
                self.conv.append(nn.Sequential(nn.Conv2d(channel, channel, 3, 1, 1, 1), nn.LeakyReLU(0.2)))
            for i in range(self.num):
                self.dense_cat(nn.Sequential(nn.Conv2d((i+2)*channel, channel, 1, 1), nn.LeakyReLU(0.2)))
    def forward(self, x):
        cat=[]
        cat.append(x)
        for i in range(self.num):
            next = self.conv[i](x)
            cat.append(next)
            x = self.dense_cat[i](torch.cat(cat,dim=1))
        return x

class DCPB(nn.Module):
    def __init__(self,in_channel,out_channel):
        super(DCPB, self).__init__()
        self.pyramid = 4
        self.scale=nn.ModuleList()
        for i in range(self.pyramid):
            self.scale.append(nn.MaxPool2d(2**i, 2**i))
        self.up=nn.ModuleList()
        for i in range(self.pyramid):
            self.up.append(nn.UpsamplingBilinear2d(scale_factor=2**i))
        self.dense = Dense_block(out_channel,3)
        self.conv=nn.ModuleList()
        for i in range(self.pyramid):
            self.conv.append(nn.Sequential(nn.Conv2d(in_channel, out_channel, 3, 1, 1, 1), nn.LeakyReLU(0.2)))
        self.cat_1 = nn.Sequential(nn.Conv2d(self.pyramid*in_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
        self.cat_2 = nn.Sequential(nn.Conv2d((self.pyramid+1)*in_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
    def forward(self, x):
        before=[]
        out=[]
        for i in range(self.pyramid):
            before.append(self.conv[i](self.scale[i](x)))
        for i in range(self.pyramid):
            out.append(self.up[i](before[i]))
        cat_1 = self.cat_1(torch.cat(out,dim=1))
        temp = self.dense(cat_1)
        out.append(temp)
        output = self.cat_2(torch.cat(out,dim=1))
        return output
class MSB(nn.Module):
    def __init__(self, in_channel, out_channel):
        super(MSB, self).__init__()
        self.num_scale = settings.scale_num
        self.scale = nn.ModuleList()
        for i in range(self.num_scale-1):
            self.scale.append(nn.MaxPool2d(2**(i+1), 2**(i+1)))
        self.res = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1),
                               nn.LeakyReLU(0.2),
                               nn.Conv2d(out_channel, out_channel, 3, 1, 1),
                               nn.LeakyReLU(0.2),
                               )
        self.cat = nn.Sequential(nn.Conv2d(self.num_scale*in_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))

    def forward(self, x):
        b, c, h, w = x.size()
        scale = []
        up = []
        up.append(x)
        for i in range(self.num_scale-1):
            scale.append(self.scale[i](x))
            up.append(F.upsample_bilinear(scale[-1], size=[h, w]))
        cat = self.cat(torch.cat(up, dim=1))
        res = self.res(cat)
        out = x+res
        return out


class MLMSB(nn.Module):
    def __init__(self, in_channel, out_channel):
        super(MLMSB, self).__init__()
        self.num_level = settings.level_num
        self.level = nn.ModuleList()
        for i in range(self.num_level):
            self.level.append(MSB(in_channel, out_channel))

    def forward(self, x):
        for i in range(self.num_level):
            x=self.level[i](x)
        return x

class EMLRB(nn.Module):
    def __init__(self,in_channel, out_channel):
        super(EMLRB, self).__init__()
        self.pyramid = 4
        self.scale=nn.ModuleList()
        for i in range(self.pyramid-1):
            self.scale.append(nn.MaxPool2d(2**(i+1), 2**(i+1)))
        self.up=nn.ModuleList()
        for i in range(self.pyramid-1):
            self.up.append(nn.UpsamplingBilinear2d(scale_factor=2**(i+1)))
        self.conv=nn.ModuleList()
        for i in range(self.pyramid):
            self.conv.append(nn.Sequential(nn.Conv2d(in_channel, out_channel, 3, 1, 1, 1), nn.LeakyReLU(0.2)))
        self.cat=nn.ModuleList()
        for i in range(self.pyramid):
            self.cat.append(nn.Sequential(nn.Conv2d((i+3)*in_channel, out_channel, 1, 1), nn.LeakyReLU(0.2)))
        self.cat_all = nn.Sequential(nn.Conv2d((self.pyramid+1)*in_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
    def forward(self,x):
        b,c,h,w=x.size()
        scale = []
        up = []
        temp_list = []
        temp=x
        temp_list.append(x)
        # for i in range(self.pyramid-1):
        #     scale.append(self.scale[i](x))
        #     up.append(self.up[i](scale[-1]))
        for i in range(self.pyramid-1):
            scale.append(self.scale[i](x))
            up.append(F.upsample_bilinear(scale[-1],size=[h,w]))

        for i in range(self.pyramid):
            conv=self.conv[i](temp)
            temp_list.append(conv)
            if i <= 2:
                temp_list.append(up[i])
                temp = self.cat[i](torch.cat(temp_list,dim=1))
                del temp_list[-1]
        output = conv+x
        return output

class ResBlock(nn.Module):
    def __init__(self,in_channel, out_channel):
        super(ResBlock, self).__init__()
        self.res = nn.Sequential(
            nn.Conv2d(in_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),
        )
    def forward(self, x):
        out = self.res(x)+x
        return out
class Coase2Fine(nn.Module):
    def __init__(self, in_channel, out_channel):
        super(Coase2Fine, self).__init__()
        self.conv7_1 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),)
        self.conv7_2 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),)
        self.conv5_1 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),
                                     )
        self.conv5_2 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),
                                     )
        self.conv3_1 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),
                                     )

        if settings.guided_in_block is True:
            self.cat_3_5 = nn.Sequential(nn.Conv2d(2 * out_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
            self.cat_5_7 = nn.Sequential(nn.Conv2d(2 * out_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
        self.cat_all = nn.Sequential(nn.Conv2d(3 * out_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
        self.scale=nn.ModuleList()
        for i in range(2):
            self.scale.append(nn.MaxPool2d(2**(i+1), 2**(i+1)))
        self.res = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1),
                                 nn.LeakyReLU(0.2),
                                 nn.Conv2d(out_channel, out_channel, 3, 1, 1),
                                 nn.LeakyReLU(0.2),
                                 )
    def forward(self,x):
        b, c, h, w = x.size()
        scale = []
        scale.append(x)
        for i in range(2):
            scale.append(self.scale[i](x))
        #  scale[2]: 1/4*1/4 ;  scale[1]: 1/2*1/2  ;  scale[0]: 1*1
        if settings.guided_in_block:
            conv3 = self.conv3_1(scale[2])
            conv5_1 = self.conv5_1(scale[1])
            b5, c5, h5, w5 = conv5_1.size()
            conv7_1 = self.conv7_1(scale[0])
            b7, c7, h7, w7 = conv7_1.size()
            conv5_2 = self.conv5_2(self.cat_3_5(torch.cat([F.upsample_bilinear(conv3, size=[h5, w5]), conv5_1], dim=1)))
            conv7_2 = self.conv7_2(self.cat_5_7(torch.cat([F.upsample_bilinear(conv5_1, size=[h7, w7]),conv7_1], dim=1)))
            cat = self.cat_all(torch.cat([F.upsample_bilinear(conv3, size=[h, w]), F.upsample_bilinear(conv5_2, size=[h, w]),conv7_2], dim=1))
            res = self.res(cat)
            output = res+x
            return output
        else:
            conv3 = self.conv3_1(scale[2])
            conv5_1 = self.conv5_1(scale[1])
            conv7_1 = self.conv7_1(scale[0])
            conv5_2 = self.conv5_2(conv5_1)
            conv7_2 = self.conv7_2(conv7_1)
            cat = self.cat_all(
                torch.cat([F.upsample_bilinear(conv3, size=[h, w]), F.upsample_bilinear(conv5_2, size=[h, w]), conv7_2],dim=1))
            res = self.res(cat)
            output = res + x

            return output


class Fine2Coase(nn.Module):
    def __init__(self, in_channel, out_channel):
        super(Fine2Coase, self).__init__()
        self.conv7_1 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),
                                     )
        self.conv5_1 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),)
        self.conv5_2 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),)
        self.conv3_1 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),
                                     )
        self.conv3_2 = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1), nn.LeakyReLU(0.2),)
        if settings.guided_in_block is True:
            self.cat_7_5 = nn.Sequential(nn.Conv2d(2 * out_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
            self.cat_5_3 = nn.Sequential(nn.Conv2d(2 * out_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
        self.cat_all = nn.Sequential(nn.Conv2d(3 * out_channel, out_channel, 1, 1), nn.LeakyReLU(0.2))
        self.scale = nn.ModuleList()
        for i in range(2):
            self.scale.append(nn.MaxPool2d(2 ** (i + 1), 2 ** (i + 1)))
        self.res = nn.Sequential(nn.Conv2d(out_channel, out_channel, 3, 1, 1),
                                 nn.LeakyReLU(0.2),
                                 nn.Conv2d(out_channel, out_channel, 3, 1, 1),
                                 nn.LeakyReLU(0.2),
                                 )

    def forward(self, x):
        b, c, h, w = x.size()
        scale = []
        scale.append(x)
        for i in range(2):
            scale.append(self.scale[i](x))
        if settings.guided_in_block is True:
            conv3_1 = self.conv3_1(scale[2])
            b3, c3, h3, w3 = conv3_1.size()
            conv5_1 = self.conv5_1(scale[1])
            b5, c5, h5, w5 = conv5_1.size()
            conv7 = self.conv7_1(scale[0])
            conv5_2 = self.conv5_2(
                self.cat_7_5(torch.cat([F.upsample_bilinear(conv7, size=[h5, w5]), conv5_1], dim=1)))
            conv3_2 = self.conv3_2(self.cat_5_3(torch.cat([F.upsample_bilinear(conv5_1, size=[h3, w3]), conv3_1], dim=1)))
            cat = self.cat_all(
                torch.cat([F.upsample_bilinear(conv3_2, size=[h, w]), F.upsample_bilinear(conv5_2, size=[h, w]), conv7],
                dim=1))
            res = self.res(cat)
            output = res + x
            return output
        else:
            conv3_1 = self.conv3_1(scale[2])
            conv5_1 = self.conv5_1(scale[1])
            conv7 = self.conv7_1(scale[0])
            conv5_2 = self.conv5_2(conv5_1)
            conv3_2 = self.conv3_2(conv3_1)
            cat = self.cat_all(
                torch.cat([F.upsample_bilinear(conv3_2, size=[h, w]), F.upsample_bilinear(conv5_2, size=[h, w]), conv7],dim=1))
            res = self.res(cat)
            output = res + x
            return output

class Coarse2fine2fine2coarse(nn.Module):
    def __init__(self, in_channel, out_channel):
        super(Coarse2fine2fine2coarse, self).__init__()
        if settings.bolck_style == 'coarse2fine' or settings.bolck_style == 'fine2coarse':
            self.coase2fine = Coase2Fine(in_channel, out_channel)
            self.fine2coase = Fine2Coase(in_channel, out_channel)
        elif settings.bolck_style == 'coarse':
            self.coase2fine = Coase2Fine(in_channel, out_channel)
        elif settings.bolck_style == 'fine':
            self.fine2coase = Fine2Coase(in_channel, out_channel)
    def forward(self, x):
        if settings.bolck_style =='coarse2fine':
            coase2fine=self.fine2coase(self.coase2fine(x))
            return coase2fine
        elif settings.bolck_style == 'fine2coarse':
            coase2fine = self.coase2fine(self.fine2coase(x))
            return coase2fine
        elif settings.bolck_style == 'coarse':
            coase2fine = self.coase2fine(x)
            return coase2fine
        elif settings.bolck_style == 'fine':
            coase2fine = self.fine2coase(x)
            return coase2fine

class MSDC(nn.Module):
    def __init__(self, in_channel,out_channel):
        super(MSDC, self).__init__()

        if settings.dilation is True:
            self.path1 = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, 3, 1, 1, 1),
                nn.LeakyReLU(0.2),
            )
            self.path3 = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, 3, 1, 3, 3),
                nn.LeakyReLU(0.2),
            )
            self.path5 = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, 3, 1, 5, 5),
                nn.LeakyReLU(0.2),
            )

            self.cat1_3 = nn.Sequential(
                nn.Conv2d(2 * out_channel, out_channel, 1, 1),
                nn.LeakyReLU(0.2)
            )
            self.cat3_5 = nn.Sequential(
                nn.Conv2d(2 * out_channel, out_channel, 1, 1),
                nn.LeakyReLU(0.2)
            )
            self.cat_final = nn.Sequential(
                nn.Conv2d(2 * out_channel, out_channel, 1, 1),
                nn.LeakyReLU(0.2),
            )
        else:
            self.path1 = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, 3, 1, 1, 1),
                nn.LeakyReLU(0.2),
            )
            self.path3 = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, 3, 1, 1, 1),
                nn.LeakyReLU(0.2),
            )
            self.path5 = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, 3, 1, 1, 1),
                nn.LeakyReLU(0.2),
            )

            self.cat1_3 = nn.Sequential(
                nn.Conv2d(2 * out_channel, out_channel, 1, 1),
                nn.LeakyReLU(0.2))
            self.cat3_5 = nn.Sequential(
                nn.Conv2d(2 * out_channel, out_channel, 1, 1),
                nn.LeakyReLU(0.2))
            self.cat_final = nn.Sequential(
                nn.Conv2d(2 * out_channel, out_channel, 1, 1),
                nn.LeakyReLU(0.2),
                )

    def forward(self, x):
        path1 = self.path1(x)
        path3 = self.path3(x)
        path5 = self.path5(x)
        cat1_3 = self.cat1_3(torch.cat([path1, path3], dim=1))
        cat3_5 = self.cat3_5(torch.cat([path3, path5], dim=1))
        final = self.cat_final(torch.cat([cat1_3, cat3_5], dim=1))
        return final+x


if settings.bolck_style == 'msdc':
    My_unit = MSDC
elif settings.bolck_style == 'dcpb':
    My_unit = DCPB
elif settings.bolck_style == 'emlrb':
    My_unit = EMLRB
elif settings.bolck_style == 'mlmsb':
    My_unit = MLMSB
else:
    My_unit = Coarse2fine2fine2coarse

class Derain_net(nn.Module):
    def __init__(self, in_channel, mid_channel, out_channel):
        super(Derain_net, self).__init__()
        self.convert = nn.Sequential(nn.Conv2d(in_channel, mid_channel, 3, 1, 1), nn.LeakyReLU(0.2),)
        self.conv_1 = nn.Sequential(My_unit(mid_channel, mid_channel),
                                    My_unit(mid_channel, mid_channel)
                                    )

        self.pooling_1 = nn.MaxPool2d(2, 2)
        self.conv_2 = nn.Sequential(My_unit(mid_channel, mid_channel),
                                    My_unit(mid_channel, mid_channel)
                                    )
        self.pooling_2 = nn.MaxPool2d(2, 2)

        self.conv_3 = RecUnit(mid_channel, mid_channel, 3, 1)

        self.conv_4 = nn.Sequential(My_unit(mid_channel, mid_channel),
                                    My_unit(mid_channel, mid_channel)
                                    )
        self.conv_5 = nn.Sequential(My_unit(mid_channel, mid_channel),
                                    My_unit(mid_channel, mid_channel)
                                    )

        self.out = nn.Sequential(nn.Conv2d(mid_channel, mid_channel, 3, 1, 1), nn.LeakyReLU(0.2),
                                 nn.Conv2d(mid_channel, out_channel, 1, 1))
        self.cat_1 = nn.Sequential(nn.Conv2d(2*mid_channel, mid_channel, 1, 1), nn.LeakyReLU(0.2))
        self.cat_2 = nn.Sequential(nn.Conv2d(2*mid_channel, mid_channel, 1, 1), nn.LeakyReLU(0.2))

    def forward(self, x, rnn_input = None):
        convert = self.convert(x)
        conv1 = self.conv_1(convert)
        b1, c1, h1, w1 = conv1.size()
        pooling1 = self.pooling_1(conv1)
        conv2 = self.conv_2(pooling1)
        b2, c2, h2, w2 = conv2.size()
        pooling2 = self.pooling_2(conv2)

        conv3, rnn = self.conv_3(pooling2, rnn_input)

        conv4 = self.conv_4(self.cat_1(torch.cat([F.upsample_bilinear(conv3, size=[h2, w2]), conv2], dim=1)))
        conv5 = self.conv_5(self.cat_2(torch.cat([F.upsample_bilinear(conv4, size=[h1, w1]), conv1], dim=1)))
        out = self.out(conv5)

        return x-out, rnn

class Derain_rnn(nn.Module):
    def __init__(self):
        super(Derain_rnn, self).__init__()
        self.in_channel = 3
        self.mid_channel = settings.channel_derain
        self.num = settings.stage_num
        self.res = nn.ModuleList()
        self.cat_1 = nn.ModuleList()
        self.cat_2 = nn.ModuleList()
        self.cat_dense=nn.ModuleList()
        if settings.guide is True and self.num > 1:
            for i in range(self.num-1):
                self.cat_dense.append(nn.Conv2d((i+2)*3, 3, 1, 1))
        if settings.shared is True:
            self.derain_list = Derain_net(3, self.mid_channel, 3)
        else:
            self.derain_list = nn.ModuleList([Derain_net(3, self.mid_channel, 3) for _ in range(self.num)])

    def forward(self, next_img):
        derain = []
        derain.append(next_img)
        rnn = None
        if settings.shared is True:
            for i in range(self.num):
                eval_derain, rnn = self.derain_list(next_img, rnn)
                derain.append(eval_derain)
                if settings.guide is True and self.num > 1 and i < self.num-1:
                    next_img = self.cat_dense[i](torch.cat(derain, dim=1))
                else:
                    next_img = derain[-1]
            return derain
        else:
            for i in range(self.num):
                eval_derain, rnn = self.dehaze_list[i](next_img, rnn)
                derain.append(eval_derain)
                if settings.guide is True:
                    next_img = self.cat_dense[i](torch.cat(derain, dim=1))
                else:
                    next_img = derain[-1]
            return derain

class Res(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(Res, self).__init__()
        self.in_dim = in_ch
        self.out_dim = out_ch
        self.res = nn.Sequential(
            nn.Conv2d(self.in_dim, self.out_dim, 3, 1, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(self.out_dim, self.out_dim, 3, 1, 1), nn.LeakyReLU(0.2),
            )
        if in_ch!=out_ch:
            self.convert=nn.Conv2d(self.in_dim,self.out_dim,1,1)

    def forward(self, x):
        if self.in_dim != self.out_dim:
            y = self.convert(x)
        else:
            y=x
        res=self.res(x)
        out=res+y
        return out


class Refinement(nn.Module):
    def __init__(self):
        super(Refinement, self).__init__()
        self.channel=settings.channel_dehaze
        self.refinement = nn.Sequential(
            Res(3,self.channel),
            Res(self.channel,self.channel),
            Res(self.channel, self.channel), Res(self.channel, 3),

        )

    def forward(self, x):
        out = self.refinement(x)
        return out


class RESCAN(nn.Module):
    def __init__(self):
        super(RESCAN, self).__init__()
        # self.extract = nn.Sequential(
        #     nn.Conv2d(3, 3, 3, 1, 1),
        #     nn.LeakyReLU(0.2),
        # )
        self.net = Derain_rnn()
        if settings.refinement is True:
            self.refine = Refinement()

    def forward(self, x):
        # extract = self.extract(x)
        derain = self.net(x)   #rain是一个列表
        if settings.refinement is True:
            refine = self.refine(x)
            derain.append(refine)
        del derain[0]
        return derain

class VGG(nn.Module):
    'Pretrained VGG-19 model features.'
    def __init__(self, layers=(3), replace_pooling = False):
        super(VGG, self).__init__()
        self.layers = layers
        self.instance_normalization = nn.InstanceNorm2d(128)
        self.relu = nn.ReLU()
        self.model = vgg19(pretrained=True).features
        # Changing Max Pooling to Average Pooling
        if replace_pooling:
            self.model._modules['4'] = nn.AvgPool2d((2,2), (2,2), (1,1))
            self.model._modules['9'] = nn.AvgPool2d((2,2), (2,2), (1,1))
            self.model._modules['18'] =nn.AvgPool2d((2,2), (2,2), (1,1))
            self.model._modules['27'] =nn.AvgPool2d((2,2), (2,2), (1,1))
            self.model._modules['36'] = nn.AvgPool2d((2,2), (2,2), (1,1))
        for param in self.model.parameters():
            param.requires_grad = False

    def forward(self, x):
        features = []
        for name, layer in enumerate(self.model):
            x = layer(x)
            # if name in self.layers:
            #     features.append(x)
            #     if len(features) == len(self.layers):
            #         break
        return x

#ts = torch.Tensor(16, 3, 64, 64).cuda()
#vr = Variable(ts)
#net = RESCAN().cuda()
#print(net)
#oups = net(vr)
#for oup in oups:
#    print(oup.size())
